#pragma once
#ifndef DECIMALNATIVE
#define DECIMALNATIVE

#ifdef __cplusplus
#include <cstdint>
#else
#include <stdint.h>
#endif

#define DECIMALNATIVE_VERSION "1.0.2-SNAPSHOT"

#ifdef __cplusplus
#    define DECIMALNATIVE_MANGLING extern "C"
#else
#    define DECIMALNATIVE_MANGLING extern
#endif

#ifdef DECIMALNATIVE_SHARED
#    ifdef _MSC_VER
#        define DECIMALNATIVE_EXPORT    __declspec(dllimport)
#    else
#        define DECIMALNATIVE_EXPORT    __attribute__ ((visibility("default")))
#    endif
#else
#    define DECIMALNATIVE_EXPORT
#endif

#define DECIMALNATIVE_API(x) DECIMALNATIVE_MANGLING DECIMALNATIVE_EXPORT x

typedef struct {
    uint64_t val;
} decimal_native_t;

inline static uint64_t ddfp_toUnderlying(decimal_native_t _value) {
    return _value.val;
}

inline static decimal_native_t ddfp_fromUnderlying(uint64_t value) {
    decimal_native_t dn;
    dn.val = value;
    return dn;
}

static const uint64_t DECIMAL_NATIVE_UNDERLYING_POSITIVE_INFINITY = 0x7800000000000000ULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_NEGATIVE_INFINITY = 0xF800000000000000ULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_NAN = 0x7C00000000000000ULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_NULL = 0xFFFFFFFFFFFFFF80ULL;	// = -0x80

static const uint64_t DECIMAL_NATIVE_UNDERLYING_MIN_VALUE = 0xF7FB86F26FC0FFFFULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_MAX_VALUE = 0x77FB86F26FC0FFFFULL;

static const uint64_t DECIMAL_NATIVE_UNDERLYING_MIN_POSITIVE_VALUE = 0x0000000000000001ULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_MAX_NEGATIVE_VALUE = 0x8000000000000001ULL;

static const uint64_t DECIMAL_NATIVE_UNDERLYING_ZERO = 0x31C0000000000000ULL; // e=0, m=0, sign=0
static const uint64_t DECIMAL_NATIVE_UNDERLYING_ONE = 0x31C0000000000001ULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_TWO = 0x31C0000000000002ULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_TEN = 0x31C000000000000AULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_HUNDRED = 0x31C0000000000064ULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_THOUSAND = 0x31C00000000003E8ULL;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_MILLION = 0x31C00000000F4240ULL;

static const uint64_t DECIMAL_NATIVE_UNDERLYING_ONETENTH = 0x31A0000000000000ULL + 1;
static const uint64_t DECIMAL_NATIVE_UNDERLYING_ONEHUNDREDTH = 0x3180000000000000ULL + 1;

DECIMALNATIVE_API(decimal_native_t) ddfp_fromFloat64(double x);
DECIMALNATIVE_API(decimal_native_t) ddfp_fromFloat32(float x);
DECIMALNATIVE_API(decimal_native_t) ddfp_fromInt64(int64_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_fromUInt64(uint64_t x);
DECIMALNATIVE_API(double) ddfp_toFloat64(decimal_native_t x);
DECIMALNATIVE_API(float) ddfp_toFloat32(decimal_native_t x);
DECIMALNATIVE_API(int64_t) ddfp_toInt64(decimal_native_t x);
DECIMALNATIVE_API(uint64_t) ddfp_toUInt64(decimal_native_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_fromFixedPoint64(int64_t mantissa, int32_t tenPowerFactor);
DECIMALNATIVE_API(int64_t) ddfp_toFixedPoint(decimal_native_t value, int32_t numberOfDigits);
DECIMALNATIVE_API(decimal_native_t) ddfp_parse(const char * str);
DECIMALNATIVE_API(decimal_native_t) ddfp_tryParse(const char * str, uint32_t * exception);
DECIMALNATIVE_API(const char *) ddfp_to_string(decimal_native_t x);
DECIMALNATIVE_API(const char *) ddfp_to_string_2(decimal_native_t x, char decimalMark);
DECIMALNATIVE_API(const char *) ddfp_to_string_3(decimal_native_t x, char decimalMark, char * buffer512);
DECIMALNATIVE_API(const char *) ddfp_to_scientific_string(decimal_native_t x);
DECIMALNATIVE_API(const char *) ddfp_to_scientific_string_2(decimal_native_t x, char decimalMark);
DECIMALNATIVE_API(const char *) ddfp_to_scientific_string_3(decimal_native_t x, char decimalMark, char * buffer512);
DECIMALNATIVE_API(int) ddfp_isNaN(decimal_native_t x);
DECIMALNATIVE_API(int) ddfp_isInfinity(decimal_native_t x);
DECIMALNATIVE_API(int) ddfp_isPositiveInfinity(decimal_native_t x);
DECIMALNATIVE_API(int) ddfp_isNegativeInfinity(decimal_native_t x);
DECIMALNATIVE_API(int) ddfp_isFinite(decimal_native_t x);
DECIMALNATIVE_API(int) ddfp_isNormal(decimal_native_t x);
DECIMALNATIVE_API(int) ddfp_signBit(decimal_native_t x);
DECIMALNATIVE_API(int32_t) ddfp_compare(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(int) ddfp_isEqual(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(int) ddfp_isNotEqual(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(int) ddfp_isLess(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(int) ddfp_isLessOrEqual(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(int) ddfp_isGreater(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(int) ddfp_isGreaterOrEqual(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(int) ddfp_isZero(decimal_native_t a);
DECIMALNATIVE_API(int) ddfp_isNonZero(decimal_native_t a);
DECIMALNATIVE_API(int) ddfp_isPositive(decimal_native_t a);
DECIMALNATIVE_API(int) ddfp_isNegative(decimal_native_t a);
DECIMALNATIVE_API(int) ddfp_isNonPositive(decimal_native_t a);
DECIMALNATIVE_API(int) ddfp_isNonNegative(decimal_native_t a);
DECIMALNATIVE_API(decimal_native_t) ddfp_roundTowardsPositiveInfinity(decimal_native_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_roundTowardsNegativeInfinity(decimal_native_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_roundTowardsZero(decimal_native_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_roundToNearestTiesAwayFromZero(decimal_native_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_roundToNearestTiesToEven(decimal_native_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_max2(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(decimal_native_t) ddfp_max3(decimal_native_t a, decimal_native_t b, decimal_native_t c);
DECIMALNATIVE_API(decimal_native_t) ddfp_max4(decimal_native_t a, decimal_native_t b, decimal_native_t c, decimal_native_t d);
DECIMALNATIVE_API(decimal_native_t) ddfp_min2(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(decimal_native_t) ddfp_min3(decimal_native_t a, decimal_native_t b, decimal_native_t c);
DECIMALNATIVE_API(decimal_native_t) ddfp_min4(decimal_native_t a, decimal_native_t b, decimal_native_t c, decimal_native_t d);
DECIMALNATIVE_API(decimal_native_t) ddfp_negate(decimal_native_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_abs(decimal_native_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_add2(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(decimal_native_t) ddfp_add3(decimal_native_t a, decimal_native_t b, decimal_native_t c);
DECIMALNATIVE_API(decimal_native_t) ddfp_add4(decimal_native_t a, decimal_native_t b, decimal_native_t c, decimal_native_t d);
DECIMALNATIVE_API(decimal_native_t) ddfp_subtract(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(decimal_native_t) ddfp_multiply2(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(decimal_native_t) ddfp_multiply3(decimal_native_t a, decimal_native_t b, decimal_native_t c);
DECIMALNATIVE_API(decimal_native_t) ddfp_multiply4(decimal_native_t a, decimal_native_t b, decimal_native_t c, decimal_native_t d);
DECIMALNATIVE_API(decimal_native_t) ddfp_multiplyByInt32(decimal_native_t a, int32_t integer);
DECIMALNATIVE_API(decimal_native_t) ddfp_multiplyByInt64(decimal_native_t a, int64_t integer);
DECIMALNATIVE_API(decimal_native_t) ddfp_divide(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(decimal_native_t) ddfp_divideByInt32(decimal_native_t x, int32_t integer);
DECIMALNATIVE_API(decimal_native_t) ddfp_divideByInt64(decimal_native_t x, int64_t integer);
DECIMALNATIVE_API(decimal_native_t) ddfp_multiplyAndAdd(decimal_native_t a, decimal_native_t b, decimal_native_t c);
DECIMALNATIVE_API(decimal_native_t) ddfp_scaleByPowerOfTen(decimal_native_t a, int32_t n);
DECIMALNATIVE_API(decimal_native_t) ddfp_mean2(decimal_native_t a, decimal_native_t b);
DECIMALNATIVE_API(decimal_native_t) ddfp_nextUp(decimal_native_t x);
DECIMALNATIVE_API(decimal_native_t) ddfp_nextDown(decimal_native_t x);

#endif
